@@ -584,21 +584,19 @@ set_proxy_env (GHashTable *env_table,
 }
 
 static void
-setup_proxy_env (GSettings  *proxy_settings,
-                 const char *child_schema_id,
+setup_proxy_env (TerminalApp* app,
+                 TerminalProxyProtocol protocol,
                  const char *proxy_scheme,
                  const char *env_name,
                  GHashTable *env_table)
 {
-  gs_unref_object GSettings *child_settings;
   GString *buf;
   gs_free char *host;
   int port;
-  gboolean is_http;
 
-  is_http = (strcmp (child_schema_id, "http") == 0);
+  gboolean is_http = (protocol == TERMINAL_PROXY_HTTP);
 
-  child_settings = g_settings_get_child (proxy_settings, child_schema_id);
+  GSettings *child_settings = terminal_app_get_proxy_settings_for_protocol(app, protocol);
 
   host = g_settings_get_string (child_settings, "host");
   port = g_settings_get_int (child_settings, "port");
@@ -635,23 +633,6 @@ setup_proxy_env (GSettings  *proxy_settings,
   set_proxy_env (env_table, env_name, g_string_free (buf, FALSE));
 }
 
-static void
-setup_autoconfig_proxy_env (GSettings *proxy_settings,
-                            GHashTable *env_table)
-{
-  /* XXX  Not sure what to do with this.  See bug #596688.
-  gs_free char *url;
-
-  url = g_settings_get_string (proxy_settings, "autoconfig-url");
-  if (url[0])
-    {
-      char *proxy;
-      proxy = g_strdup_printf ("pac+%s", url);
-      set_proxy_env (env_table, "http_proxy", proxy);
-    }
-  */
-}
-
 static void
 setup_ignore_proxy_env (GSettings *proxy_settings,
                         GHashTable *env_table)
@@ -684,25 +665,25 @@ setup_ignore_proxy_env (GSettings *proxy_settings,
 void
 terminal_util_add_proxy_env (GHashTable *env_table)
 {
-  GSettings *proxy_settings;
+  TerminalApp* app = terminal_app_get();
   GDesktopProxyMode mode;
 
-  proxy_settings = terminal_app_get_proxy_settings (terminal_app_get ());
+  GSettings* proxy_settings = terminal_app_get_proxy_settings(app);
   mode = g_settings_get_enum (proxy_settings, "mode");
 
   if (mode == G_DESKTOP_PROXY_MODE_MANUAL)
     {
-      setup_proxy_env (proxy_settings, "http", "http", "http_proxy", env_table);
+      setup_proxy_env (app, TERMINAL_PROXY_HTTP, "http", "http_proxy", env_table);
       /* Even though it's https, the proxy scheme is 'http'. See bug #624440. */
-      setup_proxy_env (proxy_settings, "https", "http", "https_proxy", env_table);
+      setup_proxy_env (app, TERMINAL_PROXY_HTTPS, "http", "https_proxy", env_table);
       /* Even though it's ftp, the proxy scheme is 'http'. See bug #624440. */
-      setup_proxy_env (proxy_settings, "ftp", "http", "ftp_proxy", env_table);
-      setup_proxy_env (proxy_settings, "socks", "socks", "all_proxy", env_table);
+      setup_proxy_env (app, TERMINAL_PROXY_FTP, "http", "ftp_proxy", env_table);
+      setup_proxy_env (app, TERMINAL_PROXY_SOCKS, "socks", "all_proxy", env_table);
       setup_ignore_proxy_env (proxy_settings, env_table);
     }
   else if (mode == G_DESKTOP_PROXY_MODE_AUTO)
     {
-      setup_autoconfig_proxy_env (proxy_settings, env_table);
+      /* Not supported */
     }
 }
 
@@ -794,46 +775,6 @@ s_to_rgba (GVariant *variant,
   return TRUE;
 }
 
-/**
- * terminal_g_settings_new:
- * @schema_id: a settings schema ID
- * @mandatory_key: the name of a key that must exist in the schema
- * @mandatory_key_type: the expected value type of @mandatory_key
- *
- * Creates a #GSettings for @schema_id, if this schema exists and
- * has a key named @mandatory_key (if non-%NULL) with the value type
- * @mandatory_key_type.
- *
- * Returns: (transfer full): a new #GSettings, or %NULL
- */
-GSettings *
-terminal_g_settings_new (const char *schema_id,
-                         const char *mandatory_key,
-                         const GVariantType *mandatory_key_type)
-{
-  gs_unref_settings_schema GSettingsSchema *schema;
-
-  schema = g_settings_schema_source_lookup (g_settings_schema_source_get_default (),
-                                            schema_id,
-                                            TRUE);
-  if (schema == NULL)
-    return NULL;
-
-  if (mandatory_key) {
-    gs_unref_settings_schema_key GSettingsSchemaKey *key;
-
-    key = g_settings_schema_get_key (schema, mandatory_key);
-    if (key == NULL)
-      return NULL;
-
-    if (!g_variant_type_equal (g_settings_schema_key_get_value_type (key),
-                               mandatory_key_type))
-      return NULL;
-  }
-
-  return g_settings_new_full (schema, NULL, NULL);
-}
-
 /**
  * terminal_g_settings_get_rgba:
  * @settings: a #GSettings
@@ -1442,8 +1383,10 @@ terminal_util_save_print_settings (GtkPrintSettings *settings,
 const char*
 terminal_util_translate_encoding (const char *encoding)
 {
+  G_GNUC_BEGIN_IGNORE_DEPRECATIONS;
   if (vte_get_encoding_supported (encoding))
     return encoding;
+  G_GNUC_END_IGNORE_DEPRECATIONS;
 
   /* ICU knows (or has aliases for) most of the old names, except the following */
   struct {
@@ -1622,3 +1565,382 @@ terminal_util_check_envv(char const* const* strv)
 
   return TRUE;
 }
+
+#define TERMINAL_SCHEMA_VERIFIER_ERROR (g_quark_from_static_string("TerminalSchemaVerifier"))
+
+typedef enum {
+  TERMINAL_SCHEMA_VERIFIER_SCHEMA_MISSING,
+  TERMINAL_SCHEMA_VERIFIER_SCHEMA_PATH,
+  TERMINAL_SCHEMA_VERIFIER_KEY_MISSING,
+  TERMINAL_SCHEMA_VERIFIER_KEY_TYPE,
+  TERMINAL_SCHEMA_VERIFIER_KEY_DEFAULT,
+  TERMINAL_SCHEMA_VERIFIER_KEY_RANGE_TYPE,
+  TERMINAL_SCHEMA_VERIFIER_KEY_RANGE,
+  TERMINAL_SCHEMA_VERIFIER_KEY_RANGE_TYPE_UNKNOWN,
+  TERMINAL_SCHEMA_VERIFIER_KEY_RANGE_TYPE_MISMATCH,
+  TERMINAL_SCHEMA_VERIFIER_KEY_RANGE_ENUM_VALUE,
+  TERMINAL_SCHEMA_VERIFIER_KEY_RANGE_INTERVAL,
+  TERMINAL_SCHEMA_VERIFIER_CHILD_MISSING,
+} TerminalSchemaVerifierError;
+
+static gboolean
+strv_contains(char const* const* strv,
+              char const* str)
+{
+  if (strv == NULL)
+    return FALSE;
+
+  for (size_t i = 0; strv[i]; i++) {
+    if (g_str_equal (strv[i], str))
+      return TRUE;
+  }
+
+  return FALSE;
+}
+
+static gboolean
+schema_key_range_compatible(GSettingsSchema* source_schema,
+                            GSettingsSchemaKey* source_key,
+                            char const* key,
+                            GSettingsSchemaKey* reference_key,
+                            GError** error)
+{
+  gs_unref_variant GVariant* source_range =
+    g_settings_schema_key_get_range(source_key);
+  gs_unref_variant GVariant* reference_range =
+    g_settings_schema_key_get_range(reference_key);
+
+  char const* source_type = NULL;
+  gs_unref_variant GVariant* source_data = NULL;
+  g_variant_get(source_range, "(&sv)", &source_type, &source_data);
+
+  char const* reference_type = NULL;
+  gs_unref_variant GVariant* reference_data = NULL;
+  g_variant_get(reference_range, "(&sv)", &reference_type, &reference_data);
+
+  if (!g_str_equal(source_type, reference_type)) {
+    g_set_error(error, TERMINAL_SCHEMA_VERIFIER_ERROR,
+                TERMINAL_SCHEMA_VERIFIER_KEY_RANGE_TYPE,
+                "Schema \"%s\" key \"%s\" has range type \"%s\" but reference range type is \"%s\"",
+                g_settings_schema_get_id(source_schema),
+                key, source_type, reference_type);
+    return FALSE;
+  }
+
+  if (g_str_equal(reference_type, "type"))
+    ; /* no constraints; this is fine */
+  else if (g_str_equal(reference_type, "enum")) {
+    size_t source_values_len = 0;
+    gs_free char const** source_values = g_variant_get_strv(source_data, &source_values_len);
+
+    size_t reference_values_len = 0;
+    gs_free char const** reference_values = g_variant_get_strv(reference_data, &reference_values_len);
+
+    /* Check that every enum value in source is valid according to the reference */
+    for (size_t i = 0; i < source_values_len; ++i) {
+      if (!strv_contains(reference_values, source_values[i])) {
+        g_set_error(error, TERMINAL_SCHEMA_VERIFIER_ERROR,
+                    TERMINAL_SCHEMA_VERIFIER_KEY_RANGE_ENUM_VALUE,
+                    "Schema \"%s\" key \"%s\" has enum value \"%s\" not in reference schema",
+                    g_settings_schema_get_id(source_schema),
+                    key, source_values[i]);
+        return FALSE;
+      }
+    }
+  } else if (g_str_equal(reference_type, "flags")) {
+    /* Our schemas don't use flags. If that changes, need to implement this! */
+    g_assert_not_reached();
+  } else if (g_str_equal(reference_type, "range")) {
+    if (!g_variant_is_of_type(source_data,
+                              g_variant_get_type(reference_data))) {
+      char const* source_type_str = g_variant_get_type_string(source_data);
+      char const* reference_type_str = g_variant_get_type_string(reference_data);
+      g_set_error(error, TERMINAL_SCHEMA_VERIFIER_ERROR,
+                  TERMINAL_SCHEMA_VERIFIER_KEY_RANGE_TYPE_MISMATCH,
+                  "Schema \"%s\" key \"%s\" has range type \"%s\" but reference range type is \"%s\"",
+                  g_settings_schema_get_id(source_schema),
+                  key, source_type_str, reference_type_str);
+      return FALSE;
+    }
+
+    /* The source interval must be contained within the reference interval */
+    gboolean ok = FALSE;
+
+    if (g_variant_is_of_type(reference_data, G_VARIANT_TYPE("(ii)"))) {
+      int source_min, source_max;
+      g_variant_get(source_data, "(ii)", &source_min, &source_max);
+
+      int reference_min, reference_max;
+      g_variant_get(reference_data, "(ii)", &reference_min, &reference_max);
+
+      ok = (source_min >= reference_min && source_max <= reference_max);
+    } else if (g_variant_is_of_type(reference_data, G_VARIANT_TYPE("(dd)"))) {
+      double source_min, source_max;
+      g_variant_get(source_data, "(dd)", &source_min, &source_max);
+
+      double reference_min, reference_max;
+      g_variant_get(reference_data, "(dd)", &reference_min, &reference_max);
+
+      ok = (source_min >= reference_min && source_max <= reference_max);
+    } else {
+      /* Our schemas don't use this. If that changes, need to implement this! */
+      g_assert_not_reached();
+    }
+
+    if (!ok) {
+      g_set_error(error, TERMINAL_SCHEMA_VERIFIER_ERROR,
+                  TERMINAL_SCHEMA_VERIFIER_KEY_RANGE_INTERVAL,
+                  "Schema \"%s\" key \"%s\" has range interval not contained in reference range interval",
+                  g_settings_schema_get_id(source_schema), key);
+        return FALSE;
+    }
+  } else {
+    g_set_error(error, TERMINAL_SCHEMA_VERIFIER_ERROR,
+                TERMINAL_SCHEMA_VERIFIER_KEY_RANGE_TYPE_UNKNOWN,
+                "Schema \"%s\" key \"%s\" has unknown range type \"%s\"",
+                g_settings_schema_get_id(source_schema),
+                key, reference_type);
+    return FALSE;
+  }
+
+  return TRUE;
+}
+
+static gboolean
+schema_verify_key(GSettingsSchema* source_schema,
+                  char const* key,
+                  GSettingsSchema* reference_schema,
+                  GError** error)
+{
+  if (!g_settings_schema_has_key(source_schema, key)) {
+    g_set_error(error, TERMINAL_SCHEMA_VERIFIER_ERROR,
+                TERMINAL_SCHEMA_VERIFIER_KEY_MISSING,
+                "Schema \"%s\" has missing key \"%s\"",
+                g_settings_schema_get_id(source_schema), key);
+    return FALSE;
+  }
+
+  gs_unref_settings_schema_key GSettingsSchemaKey* source_key =
+    g_settings_schema_get_key(source_schema, key);
+  g_assert_nonnull(source_key);
+
+  gs_unref_settings_schema_key GSettingsSchemaKey* reference_key =
+    g_settings_schema_get_key(reference_schema, key);
+  g_assert_nonnull(reference_key);
+
+  GVariantType const* source_type = g_settings_schema_key_get_value_type(source_key);
+  GVariantType const* reference_type = g_settings_schema_key_get_value_type(reference_key);
+  if (!g_variant_type_equal(source_type, reference_type)) {
+    gs_free char* source_type_str = g_variant_type_dup_string(source_type);
+    gs_free char* reference_type_str = g_variant_type_dup_string(reference_type);
+    g_set_error(error, TERMINAL_SCHEMA_VERIFIER_ERROR,
+                TERMINAL_SCHEMA_VERIFIER_KEY_TYPE,
+                "Schema \"%s\" has type \"%s\" but reference type is \"%s\"",
+                g_settings_schema_get_id(source_schema),
+                source_type_str, reference_type_str);
+    return FALSE;
+  }
+
+  gs_unref_variant GVariant* source_default = g_settings_schema_key_get_default_value(source_key);
+  if (!g_settings_schema_key_range_check(reference_key, source_default)) {
+    gs_free char* source_value_str = g_variant_print(source_default, TRUE);
+    g_set_error(error, TERMINAL_SCHEMA_VERIFIER_ERROR,
+                TERMINAL_SCHEMA_VERIFIER_KEY_DEFAULT,
+                "Schema \"%s\" default value \"%s\" does not conform to reference schema",
+                g_settings_schema_get_id(source_schema), source_value_str);
+    return FALSE;
+  }
+
+  if (!schema_key_range_compatible(source_schema,
+                                   source_key,
+                                   key,
+                                   reference_key,
+                                   error))
+    return FALSE;
+
+  return TRUE;
+}
+
+static gboolean
+schema_verify_child(GSettingsSchema* source_schema,
+                    char const* child_name,
+                    GSettingsSchema* reference_schema,
+                    GError** error)
+{
+  /* Should verify the child's schema ID is as expected and exists in
+   * the source, but there appears to be no API to get the schema ID of
+   * the child.
+   *
+   * We work around this missing verification by never calling
+   * g_settings_get_child() and instead always constructing the child
+   * GSettings directly; and the existence and correctness of that
+   * schema is verified by the per-schema checks.
+   */
+
+  return TRUE;
+}
+
+static gboolean
+schema_verify(GSettingsSchema* source_schema,
+              GSettingsSchema* reference_schema,
+              GError** error)
+{
+  /* Verify path */
+  char const* source_path = g_settings_schema_get_path(source_schema);
+  char const* reference_path = g_settings_schema_get_path(reference_schema);
+  if (g_strcmp0(source_path, reference_path) != 0) {
+    g_set_error(error, TERMINAL_SCHEMA_VERIFIER_ERROR,
+                TERMINAL_SCHEMA_VERIFIER_SCHEMA_PATH,
+                "Schema \"%s\" has path \"%s\" but reference path is \"%s\"",
+                g_settings_schema_get_id(source_schema),
+                source_path ? source_path : "(null)",
+                reference_path ? reference_path : "(null)");
+    return FALSE;
+  }
+
+  /* Verify keys */
+  gs_strfreev char** keys = g_settings_schema_list_keys(reference_schema);
+  if (keys) {
+    for (int i = 0; keys[i]; ++i) {
+      if (!schema_verify_key(source_schema,
+                             keys[i],
+                             reference_schema,
+                             error))
+        return FALSE;
+    }
+  }
+
+  /* Verify child schemas */
+  gs_strfreev char** source_children = g_settings_schema_list_children(source_schema);
+  gs_strfreev char** reference_children = g_settings_schema_list_children(reference_schema);
+  if (reference_children) {
+    for (size_t i = 0; reference_children[i]; ++i) {
+      if (!strv_contains((char const* const*)source_children, reference_children[i])) {
+        g_set_error(error, TERMINAL_SCHEMA_VERIFIER_ERROR,
+                    TERMINAL_SCHEMA_VERIFIER_CHILD_MISSING,
+                    "Schema \"%s\" has missing child \"%s\"",
+                    g_settings_schema_get_id(source_schema),
+                    reference_children[i]);
+        return FALSE;
+      }
+
+      if (!schema_verify_child(source_schema,
+                               reference_children[i],
+                               reference_schema,
+                               error))
+          return FALSE;
+    }
+  }
+
+  return TRUE;
+}
+
+static gboolean
+schemas_source_verify_schema_by_name(GSettingsSchemaSource* source,
+                                     char const* schema_name,
+                                     GSettingsSchemaSource* reference_source,
+                                     GError** error)
+{
+  gs_unref_settings_schema GSettingsSchema* source_schema =
+    g_settings_schema_source_lookup(source, schema_name, TRUE /* recursive */);
+
+  if (!source_schema) {
+    g_set_error(error, TERMINAL_SCHEMA_VERIFIER_ERROR,
+                TERMINAL_SCHEMA_VERIFIER_SCHEMA_MISSING,
+                "Schema \"%s\" is missing", schema_name);
+    return FALSE;
+  }
+
+  gs_unref_settings_schema GSettingsSchema* reference_schema =
+    g_settings_schema_source_lookup(reference_source,
+                                    schema_name,
+                                    FALSE /* recursive */);
+  g_assert_nonnull(reference_schema);
+
+  return schema_verify(source_schema,
+                       reference_schema,
+                       error);
+}
+
+static gboolean
+schemas_source_verify_schemas(GSettingsSchemaSource* source,
+                              char const* const* schemas,
+                              GSettingsSchemaSource* reference_source,
+                              GError** error)
+{
+  if (!schemas)
+    return TRUE;
+
+  for (int i = 0; schemas[i]; ++i) {
+    if (!schemas_source_verify_schema_by_name(source,
+                                              schemas[i],
+                                              reference_source,
+                                              error))
+      return FALSE;
+  }
+
+  return TRUE;
+}
+
+static gboolean
+schemas_source_verify(GSettingsSchemaSource* source,
+                      GSettingsSchemaSource* reference_source,
+                      GError** error)
+{
+  gs_strfreev char** reloc_schemas = NULL;
+  gs_strfreev char** nonreloc_schemas = NULL;
+
+  g_settings_schema_source_list_schemas(reference_source,
+                                        FALSE /* recursive */,
+                                        &reloc_schemas,
+                                        &nonreloc_schemas);
+
+  if (!schemas_source_verify_schemas(source,
+                                     (char const* const*)reloc_schemas,
+                                     reference_source,
+                                     error))
+    return FALSE;
+
+  if (!schemas_source_verify_schemas(source,
+                                     (char const* const*)nonreloc_schemas,
+                                     reference_source,
+                                     error))
+    return FALSE;
+
+  return TRUE;
+}
+
+
+GSettingsSchemaSource*
+terminal_g_settings_schema_source_get_default(void)
+{
+  GSettingsSchemaSource* default_source = g_settings_schema_source_get_default();
+
+  gs_free_error GError* error = NULL;
+  GSettingsSchemaSource* reference_source =
+    g_settings_schema_source_new_from_directory(TERM_PKGLIBDIR,
+                                                NULL /* parent source */,
+                                                FALSE /* trusted */,
+                                                &error);
+  if (!reference_source)  {
+    /* Can only use the installed schemas, or abort here. */
+    g_printerr("Failed to load reference schemas: %s\n"
+               "Using unverified installed schemas.\n",
+               error->message);
+
+    g_settings_schema_source_unref(reference_source);
+    return g_settings_schema_source_ref(default_source);
+  }
+
+  if (!schemas_source_verify(default_source, reference_source, &error)) {
+    g_printerr("Installed schemas failed verification: %s\n"
+               "Falling back to built-in reference schemas.\n",
+               error->message);
+
+    return reference_source; /* transfer */
+  }
+
+  /* Installed schemas verified; use them. */
+  g_settings_schema_source_unref(reference_source);
+  return g_settings_schema_source_ref(default_source);
+}
